function [dX, constraints, signals, param] = liuDiesel2(X, U, N_ice, param)
% LIUDIESEL2(X, U, N_ice, param)
% LiU Diesel 2 - A MVEM of a CI, 6 cylinder in-line, 2450 Nm, 450 bhp
% engine.
% -------------------------------------------------------------------------
% Copyright 2018, Kristoffer Ekberg, Viktor Leek, Lars Eriksson
%
% This file is part of LiU Diesel 2.
%
% LiU Diesel 2 is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by
% the Free Software Foundation, either version 3 of the License, or any
% later version.
% LiU Diesel 2 is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
% Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with LiU Diesel 2.  If not, see <https://www.gnu.org/licenses/>.
% -------------------------------------------------------------------------
%% Model ------------------------------------------------------------------
% States
p_c   = X(1); % Pressure after compressor     [Pa]
p_im  = X(2); % Intake manifold pressure      [Pa]
p_em  = X(3); % Exhaust manifold pressure     [Pa]
w_tc  = X(4); % Turbocharger rotational speed [rad/s]

% Control signals
u_f   = U(1); % Fuel injection      [mg/cycle]
u_thr = U(2); % Throttle position   [0, 1]
u_wg  = U(3); % Wastegate position  [0, 1]

% Compressor --------------------------------------------------------------
[W_c, eta_c, T_c, P_c, c_constr] ... [kg/s, -, K, W, -]
    = compressor(param.p_amb, p_c, param.T_amb, w_tc, param.comp);

% Throttle ----------------------------------------------------------------
[W_thr, Pi_thr] = throttleMassflow(p_c, p_im, param.T_cac, u_thr, param);

% Intake manifold ---------------------------------------------------------
W_cyl = cylinderMassflow(p_im, param.T_im, N_ice, param); % [kg/s]

% Cylinder ----------------------------------------------------------------
W_f = fuelMassflow(u_f, N_ice, param); % [kg/s]
phi = fuelToAirEqRatio(W_f, W_cyl, param); % [-]
[M_ice, cylSig] = engineTorque(N_ice, p_im, p_em, u_f, phi, param); %[Nm]
[T_em, T_e] = ...
    exhaustGasTemperature(p_im, p_em, param.T_im, W_f, W_cyl,param);

% Turbine -----------------------------------------------------------------
[W_t, P_t_eta_tm, eta_tm, T_t] = ...
    turbine(w_tc, p_em, param.p_ats, T_em, param);

% Wastegate ---------------------------------------------------------------
W_wg = wastegateMassflow(p_em, param.p_ats, T_em, u_wg, param); % [kg/s]

% Temperature mix after turbine -------------------------------------------
T_ats = wastegateTurbineGasTemperatureMix(T_t, T_em, W_t, W_wg); % [K]

% Dynamics ----------------------------------------------------------------
dp_c = param.R_a .* T_c .* (W_c - W_thr) ./ param.V_c;
dpim = param.R_a .* param.T_im .* (W_thr - W_cyl) ./ param.V_im;
dpem = param.R_e .* T_em .* (W_cyl + W_f - W_t - W_wg) ./ param.V_em;
dwtc = (P_t_eta_tm - P_c) ./ (w_tc .* param.J_tc);
dX   = [dp_c; dpim; dpem; dwtc];

%% Constraints ------------------------------------------------------------
% Constraint 1: Torque limit 500 -1000 rpm
% Constraint 2: Torque limit 1000-1300 rpm
% Constraint 3: Torque limit 1300-1900 rpm
% Constraint 4: Torque limit 1900-2400 rpm
% Constraint 5: Turbine minimum efficiency
% Constraint 6: Smoke limiter
% Constraint 7: Compressor constraints
constraints = [...
    M_ice-100*(param.constr.M_max1(1) + param.constr.M_max1(2)*( ...
    (N_ice-param.constr.M_max1(3))/1000).^2 );...
    M_ice-2400;...
    M_ice-100*(param.constr.M_max2(1) + param.constr.M_max2(2)*N_ice/1000);...
    M_ice-100*(param.constr.M_max3(1) + param.constr.M_max3(2)*N_ice/1000);...
    param.constr.eta_tm_min - eta_tm;...
    (phi-1/param.constr.lambda_min);...
    c_constr];

%% Signals ----------------------------------------------------------------
signals=struct(...
    ... Compressor and Throttle
    'p_c'   ,p_c  , ...
    'W_thr'   ,W_thr  , ...
    'u_thr'   ,u_thr  , ...
    'Pi_thr'  ,Pi_thr , ...
    ... Intake manifold
    'p_im'    ,p_im   , ...
    'W_cyl'   ,W_cyl  , ...
    ... Cylinder
    'N_ice'   ,N_ice  , ...
    'M_ice'   ,M_ice  , ...
    'P_ice'   ,M_ice.*N_ice*pi/30  , ...
    'u_f'     ,u_f    , ...
    'W_f'     ,W_f    , ...
    'Lambda'  ,1./phi , ...
    'phi'     ,phi    , ...
    'M_fric'  ,cylSig.M_fric, ...
    'M_pump'  ,cylSig.M_pump, ...
    'M_ig'    ,cylSig.M_ig, ...
    'BMEP'    ,cylSig.BMEP, ...
    'IMEP'    ,cylSig.IMEP, ...
    'PMEP'    ,cylSig.PMEP, ...
    'FMEP'    ,cylSig.FMEP, ...
    'gamma_cyl',cylSig.gamma, ...
    'eta_otto' ,cylSig.eta_otto, ...
    'eta_inj' ,cylSig.eta_cal, ...
    'eta_ig' ,cylSig.eta_ig, ...
    ... Exhaust manifold
    'p_em'    ,p_em   , ...
    'T_em'    ,T_em   , ...
    'T_e'     ,T_e    , ...
    ... Turbo
    'w_t'     ,w_tc    , ...
    'N_t'     ,w_tc*30/pi/1e3, ... [krpm]
    'eta_tm'  ,eta_tm , ...
    ... Turbine
    'W_wg'    ,W_wg   , ...
    'u_wg'    ,u_wg   , ...
    'W_t'     ,W_t    , ...
    'T_turb'  ,T_t    , ...
    'T_exhaust' ,T_ats, ...
    'P_turb_eta_tm', P_t_eta_tm, ...
    'W_delta', W_t+W_wg-W_f, ...
    ... Compressor
    'Pi_c' ,p_c./param.p_amb  , ...
    'W_c'  ,W_c   , ...
    'eta_c',eta_c , ...
    'T_c'  ,T_c   , ...
    'P_c'  ,P_c     ...
    );
end